#include <stdio.h>
#include <memory.h>

#include "file.h"
#include "exception.h"
#include "fsfactory.h"

#ifndef _WIN32
	#define _vsnprintf vsnprintf
#endif

File::File()
{
	this->fs = FsFactory::GetFileSystem();
	accMode = ACC_NONE;
	un_ch = -1;
}

File::File(FileSystem* fs)
{
	this->fs = fs;
	accMode = ACC_NONE;
	un_ch = -1;
}

File::~File()
{
	TryNClose();
}

File::File(const File& f)
{
	throw Exception("Copy File object not allowed (construct)");
}

File& File::operator=(const File& f)
{
	throw Exception("Copy File object not allowed (operator=)");
}

void File::TryNClose(void)
{
	if (accMode != ACC_NONE) Close();
}

long File::FileSize(const char *filename)
{
	FileHandle* fh = fs->GetHandle(filename, FS_READ);
	if (!fh) throw Exception("File %s not found", filename);

	long l = fh->Seek(0L, FH_END);
	fs->CloseHandle(fh);

	return l;
}

bool File::FileExists(const char *filename)
{
	FileHandle* fh = fs->GetHandle(filename, FS_READ);
	if (!fh) return false;

	fs->CloseHandle(fh);
	return true;
}

void File::EnsureAccMode(int accMode)
{
	if (this->accMode != accMode) throw Exception("Incorrect access mode");
}

void File::Read(const char *filename)
{
	EnsureAccMode(ACC_NONE);

	handle = fs->GetHandle(filename, FS_READ);
	if (!handle) throw Exception("Read error in %s", filename);

	readSize = handle->Seek(0L, FH_END);
	handle->Seek(0L, FH_SET);

	currentFileSize = readSize;

	eof = (readSize == 0);
	len = 0;

	if (!eof) {
		handle->Read(buffer, FILECACHE_SIZE);
	}

	accMode = ACC_READ;
	un_ch = -1;
}

void File::Write(const char *filename)
{
	EnsureAccMode(ACC_NONE);

	handle = fs->GetHandle(filename, FS_WRITE);
	if (!handle) throw Exception("Write error in %s", filename);

	len = 0;
	accMode = ACC_WRITE;
}

void File::Append(const char *filename)
{
	throw Exception("Not implemented");
}

void File::Close(void)
{
	if (accMode != ACC_NONE)
	{
		if (accMode!=ACC_READ && len!=0) handle->Write(buffer, len);
		fs->CloseHandle(handle);
		accMode = ACC_NONE;
	} else throw Exception("File not opened");
}

BYTE File::GetBYTE(void)
{
	BYTE c;

	EnsureAccMode(ACC_READ);

	if (un_ch != -1)
	{
		c = un_ch;
		un_ch = -1;
		eof = un_eof;
		return c;
	}

	if (eof) return 0;

	if (len >= FILECACHE_SIZE)
	{
		handle->Read(buffer, FILECACHE_SIZE);
		len = 0;
	}

	readSize--;
	if (readSize <= 0) eof = true;

	return buffer[len++];
}

void File::PutBYTE(BYTE b)
{
	EnsureAccMode(ACC_WRITE);
	buffer[len++] = b;

	if (len >= FILECACHE_SIZE)
	{
		handle->Write(buffer, FILECACHE_SIZE);
		len = 0;
	}
}

void File::UnGetBYTE(BYTE b)
{
	EnsureAccMode(ACC_READ);
	if (un_ch != -1) throw Exception("No more symbols in unget queue");

	un_ch = b;
	un_eof = eof;

	eof = false;
}

bool File::Eof(void)
{
	EnsureAccMode(ACC_READ);
	return eof;
}

int File::GetC(void)
{
	return (int)GetBYTE();
}

void File::UnGetC(int c)
{
	UnGetBYTE((BYTE)c);
}

void File::PutC(int c)
{
	PutBYTE((BYTE)c);
}

void File::PutWORD(WORD w)
{
	PutBYTE((BYTE)(w & 0xFF));
	PutBYTE((BYTE)(w >> 8));
}

void File::PutDWORD(DWORD d)
{
	PutWORD((WORD)(d & (DWORD)0xFFFF));
	PutWORD((WORD)(d >> 0x10));
}

WORD File::GetWORD(void)
{
	int dummy1, dummy2;
	dummy1 = GetBYTE();
	dummy2 = GetBYTE();
	return (WORD)(dummy1 + 0x100*dummy2);	//   VisualC++6.0, -    return (WORD)(GetBYTE() + 0x100*GetBYTE())
}

DWORD File::GetDWORD(void)
{
	DWORD dummy1, dummy2;
	dummy1 = GetWORD();
	dummy2 = GetWORD();
	return (DWORD)(dummy1 + (DWORD)0x10000*dummy2);		//   VisualC++6.0, -    return (DWORD)(GetWORD() + 0x10000*(DWORD)GetWORD())
}

void File::PrintF(const char* fmt, ...)
{
	char buf[0x1000];
	int sz, i;
	va_list argptr;

	EnsureAccMode(ACC_WRITE);

	va_start(argptr, fmt);
	sz = _vsnprintf(buf, sizeof(buf), fmt, argptr);
	va_end(argptr);

	for (i = 0; i < sz; i++) PutBYTE(buf[i]);
}

void File::WriteBytes(BYTE *buf, unsigned int size)
{
	EnsureAccMode(ACC_WRITE);

	while (size)
	{
		PutBYTE(*buf);
		buf++;
		size--;
	}
}

unsigned int File::ReadBytes(BYTE *buf, unsigned int size)
{
	EnsureAccMode(ACC_READ);

	int readed = 0;

	while (size && !eof)
	{
		*(buf++) = GetBYTE();
		size--;
		readed++;
	}

	return readed;
}

void File::GetS(char *buf, unsigned int size)
{
	bool r;
	int o, c;

	EnsureAccMode(ACC_READ);

	o = 0;
	r = true;

	do
	{
		if (eof) r = false;
		else
		{
			c = GetC();

			if (c == 0x0D)
			{
				r = false;

				if (!eof)
				{
					c = GetC();
					if (c != 0x0A) UnGetC(c);
				}
			}
			else
			if (c == 0x0A)
			{
				r = false;

				if (!eof)
				{
					c = GetC();
					if (c != 0x0D) UnGetC(c);
				}
			}
			else
			{
				if (o >= size-1)
				{
					UnGetC(c);
					r = false;
				} else buf[o++] = c;
			}
		}
	} while (r);
	
	buf[o] = 0;
}

void File::SetFilePointer(long fp)
{
	EnsureAccMode(ACC_READ);

	handle->Seek(fp, FH_SET);
	readSize = currentFileSize - fp;
	eof = (readSize <= 0);
	len = FILECACHE_SIZE + 1;
}

long File::GetCurrentFileSize(void)
{
	EnsureAccMode(ACC_READ);
	return currentFileSize;
}

bool File::IsOpened(void)
{
	return (accMode != ACC_NONE);
}

SDL_RWops* File::AllocSdlRwOps(void)
{
	SDL_RWops* rwo = SDL_AllocRW();
	if (!rwo) return NULL;

	rwo->seek = SdlRwOpsSeekFunc;
	rwo->read = SdlRwOpsReadFunc;
	rwo->write = SdlRwOpsWriteFunc;
	rwo->close = SdlRwOpsCloseFunc;
	rwo->hidden.unknown.data1 = this;

	return rwo;
}

void File::FreeSdlRwOps(SDL_RWops* rwo)
{
	SDL_FreeRW(rwo);
}

int File::SdlRwOpsSeekFunc(SDL_RWops* context, int offset, int whence)
{
	File* self = (File*)context->hidden.unknown.data1;
	self->EnsureAccMode(ACC_READ);

	if (whence==SEEK_CUR && self->len!=(FILECACHE_SIZE+1))
	{
		long bufSz = self->readSize > FILECACHE_SIZE ? FILECACHE_SIZE : self->readSize;
		long realOff = self->handle->Seek(0L, FH_CUR) - bufSz + self->len;
		self->handle->Seek(realOff, FH_SET);
	}

	long off = self->handle->Seek(offset, whence);

	self->readSize = self->currentFileSize - off;
	self->eof = (self->readSize <= 0);
	self->len = FILECACHE_SIZE + 1;

	return off;
}


int File::SdlRwOpsReadFunc(SDL_RWops* context, void* ptr, int size, int maxnum)
{
	File* self = (File*)context->hidden.unknown.data1;
	return self->ReadBytes((BYTE*)ptr, size * maxnum) / size;
}

int File::SdlRwOpsWriteFunc(SDL_RWops* context, const void* ptr, int size, int num)
{
	File* self = (File*)context->hidden.unknown.data1;
	self->WriteBytes((BYTE*)ptr, size * num);
	return num;
}

int File::SdlRwOpsCloseFunc(SDL_RWops* context)
{
	File* self = (File*)context->hidden.unknown.data1;
	self->FreeSdlRwOps(context);
	return 0;
}
