#include "font.h"
#include "exception.h"
#include "utf8.h"
#include "file.h"

#include <vector>

using namespace std;

Font::Font()
{
	font = NULL;
	realHeight = 0;

	fgColor.r = 0;
	fgColor.g = 0;
	fgColor.b = 0;

	outColor.r = 255;
	outColor.g = 255;
	outColor.b = 255;
}

Font::~Font()
{
	if (font) TTF_CloseFont(font);
}

Font::Font(const Font& f)
{
	Error("Copy Font object not allowed (construct)");
}

Font& Font::operator=(const Font& f)
{
	Error("Copy Font object not allowed (operator=)");
}

void Font::Load(const char* path, unsigned int size, unsigned int renderstyle)
{
	if (font) TTF_CloseFont(font);
	if (fl.IsOpened()) fl.Close();

	fl.Read(path);
	SDL_RWops* rwo = fl.AllocSdlRwOps();
	font = TTF_OpenFontRW(rwo, 1, size);

	if (!font) throw Exception("Couldn't load %d pt font from %s: %s\n", size, path, SDL_GetError());

	TTF_SetFontStyle(font, TTF_STYLE_NORMAL | renderstyle);
	realHeight = TTF_FontHeight(font);

	for (int i = 0; i < 0x100; i++) widths[i] = -1;
}

void Font::Blit(SDL_Surface* dest, int x, int y, char c, bool outline)
{
	SDL_Surface* glyph;
	const char* ch = ConvertToUtf8(c);

	if (outline)
	{
		glyph = TTF_RenderUTF8_Blended(font, ch, outColor);

		if (glyph)
		{
			SDL_Rect d;
			d.w = glyph->w;
			d.h = glyph->h;

			d.x = x-2; d.y = y;   SDL_BlitSurface(glyph, NULL, dest, &d);
			d.x = x+2; d.y = y;   SDL_BlitSurface(glyph, NULL, dest, &d);
			d.x = x;   d.y = y-2; SDL_BlitSurface(glyph, NULL, dest, &d);
			d.x = x;   d.y = y+2; SDL_BlitSurface(glyph, NULL, dest, &d);

			d.x = x-1; d.y = y;   SDL_BlitSurface(glyph, NULL, dest, &d);
			d.x = x+1; d.y = y;   SDL_BlitSurface(glyph, NULL, dest, &d);
			d.x = x;   d.y = y-1; SDL_BlitSurface(glyph, NULL, dest, &d);
			d.x = x;   d.y = y+1; SDL_BlitSurface(glyph, NULL, dest, &d);
			d.x = x-1; d.y = y-1; SDL_BlitSurface(glyph, NULL, dest, &d);
			d.x = x+1; d.y = y-1; SDL_BlitSurface(glyph, NULL, dest, &d);
			d.x = x-1; d.y = y+1; SDL_BlitSurface(glyph, NULL, dest, &d);
			d.x = x+1; d.y = y+1; SDL_BlitSurface(glyph, NULL, dest, &d);

			SDL_FreeSurface(glyph);
		}
	}

	glyph = TTF_RenderUTF8_Blended(font, ch, fgColor);

	if (glyph)
	{
		if (widths[(unsigned char)c] == -1) widths[(unsigned char)c] = glyph->w;

		SDL_Rect d;
		d.x = x;
		d.y = y;
		d.w = glyph->w;
		d.h = glyph->h;

		SDL_BlitSurface(glyph, NULL, dest, &d);
		SDL_FreeSurface(glyph);
	}
}

void Font::BlitRects(list<SDL_Rect>& rects, SDL_Surface* dest, int x, int y, char c, bool outline)
{
	rf.Clear();
	list<SDL_Rect>::iterator it;

	SDL_Rect dr;
	dr.x = x;
	dr.y = y;
	dr.w = Width(c);
	dr.h = Height();

	for (it = rects.begin(); it != rects.end(); it++) rf.AddRect(&(*it));
	rf.BoundRect(&dr);

	int ex = dr.x + dr.w;
	int ey = dr.y + dr.h;

	vector<SDL_Rect>::iterator ri;

	for (ri = rf.rects.begin(); ri != rf.rects.end(); ri++)
	{
		int rex = ri->x + ri->w;
		int rey = ri->y + ri->h;

		if (x<=rex && ex>ri->x && y<=rey && ey>ri->y)
		{
			Blit(dest, x, y, c, outline);
			break;
		}
	}
}

// void Font::Blit(Picture& dest, int x, int y, char c)
// {
// }

int Font::Height(void)
{
	return realHeight;
}

int Font::Width(char c)
{
	if (widths[(unsigned char)c] == -1)
	{
		int res = 0;
		SDL_Color clr = {0, 0, 0};
		SDL_Surface* glyph = TTF_RenderUTF8_Solid(font, ConvertToUtf8(c), clr);

		if (glyph)
		{
			res = glyph->w;
			SDL_FreeSurface(glyph);
		}

		widths[(unsigned char)c] = res;
	}

	return widths[(unsigned char)c];
}

void Font::SetColor(int c)
{
	fgColor.r = GETR(c);
	fgColor.g = GETG(c);
	fgColor.b = GETB(c);
}

void Font::SetOutColor(int c)
{
	outColor.r = GETR(c);
	outColor.g = GETG(c);
	outColor.b = GETB(c);
}
