#include "defines.h"
#include "exception.h"
#include "global.h"

#include <SDL_ttf.h>

#include "game.h"
#include "picture.h"
#include "sound.h"
#include "thread.h"
#include "utf8.h"
#include "fsfactory.h"

#ifndef SETTINGS_NAME
	#define SETTINGS_NAME "user.settings"
#endif

SDL_Surface* realScreen;
SDL_Rect mousePrev, mouseCurr;
Picture mouseCursor[CURSOR_LAST];
SDL_Surface* mouseBuffer = NULL;

void TryRun(int val, int errVal, const char * msg)
{
	if (val == errVal) Error(msg);
}

void InitAll(void)
{
	SDL_PixelFormat* fmt = realScreen->format;
	screen = SDL_CreateRGBSurface(SDL_SWSURFACE, realScreen->w, realScreen->h, fmt->BitsPerPixel, fmt->Rmask, fmt->Gmask, fmt->Bmask, 0);

	mouseCursor[CURSOR_ARROW].Load("system/cursors/arrow.png");
	mouseCursor[CURSOR_BUSY].Load("system/cursors/busy.png");
	mouseCursor[CURSOR_HAND].Load("system/cursors/hand.png");

	mouseBuffer = SDL_CreateRGBSurface(SDL_SWSURFACE, 64, 64, fmt->BitsPerPixel, fmt->Rmask, fmt->Gmask, fmt->Bmask, 0);
	screenBuffer = SDL_CreateRGBSurface(SDL_SWSURFACE, realScreen->w, realScreen->h, fmt->BitsPerPixel, fmt->Rmask, fmt->Gmask, fmt->Bmask, 0);
	screenSecondaryBuffer = SDL_CreateRGBSurface(SDL_SWSURFACE, realScreen->w, realScreen->h, fmt->BitsPerPixel, fmt->Rmask, fmt->Gmask, fmt->Bmask, 0);

	mousePrev.w = 0;
	mousePrev.h = 0;
}

void TryFree(SDL_Surface *surf)
{
	if (surf) SDL_FreeSurface(surf);
}

void FreeAll(void)
{
	TryFree(mouseBuffer);
	TryFree(screenBuffer);
	TryFree(screen);

	TTF_Quit();
	SDL_Quit();
}

volatile bool threadEnd = false;

int GameThread(void *data)
{
	try
	{
		RunGame();
	}
	catch (Exception &e)
	{
		Error("(thread) %s", e.Message());
	}

	threadEnd = true;
	return 0;
}

SDL_Rect HideCursor(void)
{
	if (!(mousePrev.w && mousePrev.h)) return DefRect(0, 0, 0, 0);

	SDL_Rect s = DefRect(0, 0, mouseBuffer->w, mouseBuffer->h);
	SDL_Rect d = mousePrev;

	if (d.x < 0) d.x = 0;
	if (d.y < 0) d.y = 0;

	SDL_BlitSurface(mouseBuffer, &s, realScreen, &d);
	return d;
}

SDL_Rect OutputMouseCursor(void)
{
	SDL_Rect s = DefRect(0, 0, mouseBuffer->w, mouseBuffer->h);

	mousePrev.x = mouseX - 32;
	mousePrev.y = mouseY - 32;
	mousePrev.w = mouseBuffer->w;
	mousePrev.h = mouseBuffer->h;

	SDL_Rect d = mousePrev;

	if (d.x < 0) d.x = 0;
	if (d.y < 0) d.y = 0;

	SDL_BlitSurface(realScreen, &d, mouseBuffer, &s);
	mouseCursor[cursorShape].Blit(realScreen, &s, &mousePrev);

	return d;
}

#define MAX_UPDATE_RECTS 0x100

void Process()
{
	SDL_Rect rects[MAX_UPDATE_RECTS];
	SDL_Thread* thr = SDL_CreateThread(GameThread, NULL);

	while (!threadEnd)
	{
		int cnt = 0;
		rects[cnt++] = HideCursor();

		if (updateScreen)
		{
			unsigned int lockCount = 0;
			list<SDL_Rect>::iterator it;

			if (screen->locked)
			{
				lockCount = screen->locked;
				for (unsigned int i = lockCount; i--;) SDL_UnlockSurface(screen);
			}

			for (it = updateList.begin(); it != updateList.end(); it++)
			{
				SDL_BlitSurface(screen, &(*it), realScreen, &(*it));
				rects[cnt++] = *it;

				if (cnt >= MAX_UPDATE_RECTS-1) break;
			}

			if (lockCount) {
				while (lockCount--) SDL_LockSurface(screen);
			}

			updateList.clear();
			updateScreen = false;
		}

		rects[cnt++] = OutputMouseCursor();

		for (int i = 0; i < cnt; i++)
		{
			SDL_Rect *r = &rects[i];

			if (r->x + r->w < 0)
			{
				r->x = 0;
				r->w = 1;
			}
			else
			{
				if (r->x < 0) { r->w += r->x; r->x = 0; }
				if (r->x + r->w > realScreen->w) r->w = realScreen->w - r->x;
			}

			if (r->y + r->h < 0)
			{
				r->y = 0;
				r->h = 1;
			}
			else
			{
				if (r->y < 0) { r->h += r->y; r->y = 0; }
				if (r->y + r->h > realScreen->h) r->h = realScreen->h - r->y;
			}
		}

		SDL_UpdateRects(realScreen, cnt, rects);

		PollEvents();
		if (key == SDLK_ESCAPE) isActive = false;

		SDL_Delay(1);
	}

	SDL_WaitThread(thr, NULL);
}

int main(int argc, char* argv[])
{
	try
	{
		FsFactory::SetFsType(HOST_FS);
		userSettings.Load(SETTINGS_NAME);

		FsFactory::SetFsType(FOLD_FS);
		settings.Load("game.settings");
	}
	catch (Exception &e)
	{
		Error("(settings) %s", e.Message());
	}

	int width = settings.Get("Window.Width", 800);
	int height = settings.Get("Window.Height", 600);

	int enableSound = userSettings.Get("EnableSound", 1);
	int fullscreen = userSettings.Get("Fullscreen", 0);

	int cap = SDL_INIT_VIDEO;
	if (enableSound) cap |= SDL_INIT_AUDIO;

	TryRun(SDL_Init(cap), -1, "SDL_Init(cap) failed");
	TryRun(SDL_InitSubSystem(SDL_INIT_VIDEO), -1, "SDL_InitSubSystem(SDL_INIT_VIDEO) failed");

	if (enableSound) InitAudio();

	realScreen = SDL_SetVideoMode(width, height, 32, SDL_DOUBLEBUF | (fullscreen ? SDL_FULLSCREEN : 0));
	if (!realScreen) Error("SDL_SetVideoMode(width, height, 32, SDL_DOUBLEBUF | (fullscreen ? SDL_FULLSCREEN : 0)) failed");

	if (TTF_Init() < 0)
	{
		SDL_Quit();
		Error("TTF_Init() failed");
	}

	SDL_WM_SetCaption("Noname", "Noname"); 
	SDL_EnableUNICODE(1);
	SDL_ShowCursor(SDL_DISABLE);
	atexit(FreeAll);

	MakeUtf8Hash();
	
	try
	{
		channels[0].Load(string("mfx/ingame.ogg"), true);
		channels[0].SetVolume(128);

		InitAll();
		Process();
	}
	catch (Exception &e)
	{
		Error("(process) %s", e.Message());
	}

	return 0;
}
