#include "picture.h"
#include "file.h"
#include "exception.h"
#include "bmploader.h"
#include "pngloader.h"

#include <string.h>

#include <vector>

using namespace std;

Picture::Picture()
{
	width = 0;
	height = 0;
	data = NULL;
}

Picture::~Picture()
{
	if (data) delete[] data;
}

Picture::Picture(const Picture& p)
{
	if (p.width!=0 && p.height!=0)
	{
		width = p.width;
		height = p.height;
		data = new DWORD[width * height];
		memcpy(data, p.data, width * height * sizeof(DWORD));
	}
	else
	{
		width = 0;
		height = 0;
		data = NULL;
	}
}

Picture& Picture::operator=(const Picture& p)
{
	if (data) delete[] data;

	if (p.width!=0 && p.height!=0)
	{
		width = p.width;
		height = p.height;
		data = new DWORD[width * height];
		memcpy(data, p.data, width * height * sizeof(DWORD));
	}
	else
	{
		width = 0;
		height = 0;
		data = NULL;
	}

	return *this;
}

int Picture::Width(void)
{
	return width;
}

int Picture::Height(void)
{
	return height;
}

bool Picture::CheckExt(const char* path, const char* ext)
{
	int pl = strlen(path) - 1;
	int el = strlen(ext) - 1;

	if (el < 0) return true;

	for (;;)
	{
		if (pl < 0) return false;
		if (tolower(path[pl]) != tolower(ext[el])) return false;

		pl--;
		el--;

		if (el < 0) return true;
	}
}

void Picture::Clear(void)
{
	if (data) delete[] data;
	data = NULL;

	width = 0;
	height = 0;
}

void Picture::Load(const char* path, bool ignoreAlpha)
{
	if (data) delete[] data;
	data = NULL;

	File fl;
	fl.Read(path);

	if (CheckExt(path, ".bmp")) LoadBMP(&fl, &width, &height, &data, ignoreAlpha);
	else
	if (CheckExt(path, ".png")) LoadPNG(&fl, &width, &height, &data, ignoreAlpha);
	else throw Exception("Unknown image type %s", path);

	if (data == NULL) throw Exception("Error loading %s", path);

	fl.Close();
}

void Picture::Blit(SDL_Surface* dest, int x, int y, int alpha)
{
	if (!data || width<=0 || height<=0) return;

	SDL_Rect s, d;

	s.x = 0; s.y = 0; s.w = 0; s.h = 0;
	d.x = x; d.y = y; d.w = 0; d.h = 0;

	Blit(dest, &s, &d, alpha);
}

void Picture::Blit(SDL_Surface* dest, SDL_Rect *s, SDL_Rect *d, int alpha)
{
	if (!data || width<=0 || height<=0) return;

	int f_sx = s->x;
	int f_ex = (s->w ? s->x+s->w : width);
	int f_sy = s->y;
	int f_ey = (s->h ? s->y+s->h : height);

	int t_sx = d->x;
	int t_ex = (d->w ? d->x+d->w : dest->w);
	int t_sy = d->y;
	int t_ey = (d->h ? d->y+d->h : dest->h);

	if (t_sx>=dest->w || t_sy>=dest->h) return;

	if (f_ex > width) f_ex = width;
	if (f_ey > height) f_ey = height;
	if (t_ex > dest->w) t_ex = dest->w;
	if (t_ey > dest->h) t_ey = dest->h;

	if (f_ex<0 || f_ey<0 || t_ex<0 || t_ey<0) return;

	if (f_sx < 0) f_sx = 0;
	if (f_sy < 0) f_sy = 0;

	if (t_sx < 0)
	{
		f_sx -= t_sx;
		t_sx = 0;
	}

	if (t_sy < 0)
	{
		f_sy -= t_sy;
		t_sy = 0;
	}

	if (f_sx>=f_ex || f_sy>=f_ey) return;

	if (t_ex-t_sx < f_ex-f_sx) f_ex = f_sx + t_ex - t_sx;
	if (t_ey-t_sy < f_ey-f_sy) f_ey = f_sy + t_ey - t_sy;

	if (SDL_MUSTLOCK(dest)) {
		if (SDL_LockSurface(dest) < 0) return;
	}

	int tpitch = dest->pitch >> 2;

	DWORD* f = &data[f_sy * width + f_sx];
	DWORD* t = &((DWORD *)dest->pixels)[t_sy * tpitch + t_sx];

	int h = f_ey - f_sy;
	int w = f_ex - f_sx;

	tpitch -= w;

	int fpitch = width - w;

	if (alpha != 255)
	{
		while (h--)
		{
			int l = w;

			while (l--)
			{
				int fr = GET_R(*f);
				int fg = GET_G(*f);
				int fb = GET_B(*f);
				int fa = GET_A(*f);

				int tr = GET_R(*t);
				int tg = GET_G(*t);
				int tb = GET_B(*t);

				fa -= (255 - alpha);
				if (fa < 0) fa = 0;

				int oa = 0x100 - fa;
				fa++;

				int r = (fr*fa + tr*oa) >> 8;
				int g = (fg*fa + tg*oa) >> 8;
				int b = (fb*fa + tb*oa) >> 8;

				*t = DRGB(r, g, b);

				f++;
				t++;
			}

			f += fpitch;
			t += tpitch;
		}
	}
	else
	{
		while (h--)
		{
			int l = w;

			while (l--)
			{
				int fr = GET_R(*f);
				int fg = GET_G(*f);
				int fb = GET_B(*f);
				int fa = GET_A(*f);

				int tr = GET_R(*t);
				int tg = GET_G(*t);
				int tb = GET_B(*t);

				int oa = 0x100 - fa;
				fa++;

				int r = (fr*fa + tr*oa) >> 8;
				int g = (fg*fa + tg*oa) >> 8;
				int b = (fb*fa + tb*oa) >> 8;

				*t = DRGB(r, g, b);

				f++;
				t++;
			}

			f += fpitch;
			t += tpitch;
		}
	}

	if (SDL_MUSTLOCK(dest)) SDL_UnlockSurface(dest);
}

void Picture::Blit(Picture& dest, int x, int y, int alpha)
{
	if (!data || !dest.data || width<=0 || dest.width<=0 || height<=0 || dest.height<=0) return;

	SDL_Rect s, d;

	s.x = 0; s.y = 0; s.w = 0; s.h = 0;
	d.x = x; d.y = y; d.w = 0; d.h = 0;

	Blit(dest, &s, &d, alpha);
}

void Picture::Blit(Picture& dest, SDL_Rect *s, SDL_Rect *d, int alpha)
{
	if (!data || !dest.data || width<=0 || dest.width<=0 || height<=0 || dest.height<=0) return;

	int f_sx = s->x;
	int f_ex = (s->w ? s->x+s->w : width);
	int f_sy = s->y;
	int f_ey = (s->h ? s->y+s->h : height);

	int t_sx = d->x;
	int t_ex = (d->w ? d->x+d->w : dest.width);
	int t_sy = d->y;
	int t_ey = (d->h ? d->y+d->h : dest.height);

	if (t_sx>=dest.width || t_sy>=dest.height) return;

	if (f_ex > width) f_ex = width;
	if (f_ey > height) f_ey = height;
	if (t_ex > dest.width) t_ex = dest.width;
	if (t_ey > dest.height) t_ey = dest.height;

	if (f_ex<0 || f_ey<0 || t_ex<0 || t_ey<0) return;

	if (f_sx < 0) f_sx = 0;
	if (f_sy < 0) f_sy = 0;

	if (t_sx < 0)
	{
		f_sx -= t_sx;
		t_sx = 0;
	}

	if (t_sy < 0)
	{
		f_sy -= t_sy;
		t_sy = 0;
	}

	if (f_sx>=f_ex || f_sy>=f_ey) return;

	if (t_ex-t_sx < f_ex-f_sx) f_ex = f_sx + t_ex - t_sx;
	if (t_ey-t_sy < f_ey-f_sy) f_ey = f_sy + t_ey - t_sy;

	DWORD* f = &data[f_sy * width + f_sx];
	DWORD* t = &dest.data[t_sy * dest.width + t_sx];

	int h = f_ey - f_sy;
	int w = f_ex - f_sx;

	int fpitch = width - w;
	int tpitch = dest.width - w;

	if (alpha != 255)
	{
		while (h--)
		{
			int l = w;

			while (l--)
			{
				int fr = GET_R(*f);
				int fg = GET_G(*f);
				int fb = GET_B(*f);
				int fa = GET_A(*f);

				int tr = GET_R(*t);
				int tg = GET_G(*t);
				int tb = GET_B(*t);

				fa -= (255 - alpha);
				if (fa < 0) fa = 0;

				int oa = 0x100 - fa;
				fa++;

				int r = (fr*fa + tr*oa) >> 8;
				int g = (fg*fa + tg*oa) >> 8;
				int b = (fb*fa + tb*oa) >> 8;

				*t = DRGB(r, g, b);

				f++;
				t++;
			}

			f += fpitch;
			t += tpitch;
		}
	}
	else
	{
		while (h--)
		{
			int l = w;

			while (l--)
			{
				int fr = GET_R(*f);
				int fg = GET_G(*f);
				int fb = GET_B(*f);
				int fa = GET_A(*f);

				int tr = GET_R(*t);
				int tg = GET_G(*t);
				int tb = GET_B(*t);

				int oa = 0x100 - fa;
				fa++;

				int r = (fr*fa + tr*oa) >> 8;
				int g = (fg*fa + tg*oa) >> 8;
				int b = (fb*fa + tb*oa) >> 8;

				*t = DRGB(r, g, b);

				f++;
				t++;
			}

			f += fpitch;
			t += tpitch;
		}
	}
}

void Picture::BlitRects(list<SDL_Rect>& rects, SDL_Surface* dest, int x, int y, int alpha)
{
	if (!data || width<=0 || height<=0) return;

	SDL_Rect s, d;

	s.x = 0; s.y = 0; s.w = 0; s.h = 0;
	d.x = x; d.y = y; d.w = 0; d.h = 0;

	BlitRects(rects, dest, &s, &d, alpha);
}

void Picture::BlitRects(list<SDL_Rect>& rects, SDL_Surface* dest, SDL_Rect *s, SDL_Rect *d, int alpha)
{
	if (!data || width<=0 || height<=0) return;

	SDL_Rect sr = *s;
	if (!sr.w) sr.w = width - sr.x;
	if (!sr.h) sr.h = height - sr.y;

	SDL_Rect dr = *d;
	if (!dr.w) dr.w = dest->w - dr.x;
	if (!dr.h) dr.h = dest->h - dr.y;

	rf.Clear();
	list<SDL_Rect>::iterator it;

	for (it = rects.begin(); it != rects.end(); it++) rf.AddRect(&(*it));
	rf.BoundRect(&dr);

	vector<SDL_Rect>::iterator ri;

	for (ri = rf.rects.begin(); ri != rf.rects.end(); ri++)
	{
		int xoff = ri->x - dr.x;
		int yoff = ri->y - dr.y;

		if (sr.w>xoff && sr.h>yoff)
		{
			SDL_Rect r;
			r.x = sr.x + xoff;
			r.y = sr.y + yoff;
			r.w = sr.w - xoff;
			r.h = sr.h - yoff;

			Blit(dest, &r, &(*ri), alpha);
		}
	}
}

DWORD& Picture::Pixel(int x, int y)
{
	static DWORD dummy = (DWORD)0x00000000;

	if (!data || x<0 || y<0 || x>=width || y>=height)
	{
		dummy = (DWORD)0x00000000;
		return dummy;
	}

	return data[x + width * y];	
}
