#include "scenario.h"
#include "file.h"
#include "exception.h"

#include <string.h>
#include <stdlib.h>

/*
void DebugExpression(Scenario::Expression &expr, bool first=true)
{
	if (expr.type == EXPR_NUM_VALUE) AddLogN("%d", expr.numValue);
	else
	if (expr.type == EXPR_STR_VALUE)
	{
		AddLogN("\"");

		for (int i = 0; i < expr.strValue.size(); i++)
		{
			char c = expr.strValue[i];

			if (c == '"') AddLogN("\"\"");
			else if (c == '\n') AddLogN("---");
			else AddLogN("%c", c);
		}

		AddLogN("\"");
	}
	else
	if (expr.type == EXPR_OP)
	{
		switch (expr.op)
		{
			case OP_NOT:
				AddLogN("!");
				break;

			case OP_NEG:
				AddLogN("/-/");
				break;

			case OP_LTEQ:
				AddLogN("<=");
				break;

			case OP_GTEQ:
				AddLogN(">=");
				break;

			case OP_NEQ:
				AddLogN("!=");
				break;

			default:
				AddLogN("%c", expr.op);
		}

		AddLogN("(");
		bool sep = false;

		for (vector<Scenario::Expression>::iterator it = expr.params.begin(); it != expr.params.end(); it++)
		{
			if (sep) AddLogN(",");
			DebugExpression(*it, false);
			sep = true;
		}

		AddLogN(")");
	}
	else
	if (expr.type == EXPR_VAR) AddLogN("%s", expr.strValue.c_str());
	else
	if (expr.type == EXPR_FUNC)
	{
		AddLogN("%s(", expr.strValue.c_str());
		bool sep = false;

		for (vector<Scenario::Expression>::iterator it = expr.params.begin(); it != expr.params.end(); it++)
		{
			if (sep) AddLogN(", ");
			DebugExpression(*it, false);
			sep = true;
		}

		AddLogN(")");
	}
	else
	if (expr.type == EXPR_BRAC) AddLogN("[BRAC:%c]", expr.op);
	else
	if (expr.type == EXPR_SYMBOL) AddLogN("[SYMBOL:%s]", expr.strValue.c_str());
	else
	if (expr.type == EXPR_END) AddLogN("[END]");
	else AddLogN("[?:%d]", expr.type);

	if (first) AddLog("");
}
*/

Scenario::Scenario()
{
	compileTimeWarn = true;
	undefinedVariableException = false;
	Construct();
}

Scenario::Scenario(bool compileTimeWarn, bool undefinedVariableException)
{
	this->compileTimeWarn = compileTimeWarn;
	this->undefinedVariableException = undefinedVariableException;
	Construct();
}

void Scenario::Construct(void)
{
	status = SCRIPT_NONE;

	EngineFunc gotoFunc;
	gotoFunc.func = FunctionGoto;
	gotoFunc.paramNames.push_back("label");
	DefineFunc("goto", gotoFunc);

	MathFunc strMath;
	strMath.func = MathStr;
	DefineMathFunc("str", strMath);

	MathFunc randMath;
	randMath.func = MathRand;
	DefineMathFunc("rand", randMath);
}

Scenario::~Scenario()
{
	FreeScript();
}

Scenario::Scenario(const Scenario& s)
{
	throw Exception("Copy Scenario object not allowed (construct)");
}

Scenario& Scenario::operator=(const Scenario& s)
{
	throw Exception("Copy Scenario object not allowed (operator=)");
}

void Scenario::FreeScript(void)
{
	if (script.size())
	{
		vector<Line>::iterator it;

		for (it = script.begin(); it != script.end(); it++)
		{
			switch (it->type)
			{
				case OPERATOR_SET:
					delete (OperatorSet*)it->action;
					break;

				case OPERATOR_IF:
					delete (OperatorIf*)it->action;
					break;

				case OPERATOR_EXEC:
					delete (OperatorExec*)it->action;
					break;
			}
		}

		script.clear();
	}

	labels.clear();
	status = SCRIPT_NONE;
}

bool Scenario::IsNum(char c)
{
	return (c>='0' && c<='9');
}

bool Scenario::IsAlpha(char c)
{
	return ((c>='a' && c<='z') || (c>='A' && c<='Z'));
}

bool Scenario::IsAlphaNum(char c)
{
	return ((c>='0' && c<='9') || (c>='a' && c<='z') || (c>='A' && c<='Z') || c=='_');
}

void Scenario::SkipSpaces(void)
{
	while (*ptr && (*ptr==' ' || *ptr==9)) ptr++;
}

int Scenario::ReadNumber(void)
{
	int res = 0;
	if (!(*ptr)) return res;

	if (*ptr == '-')
	{
		ptr++;

		if (!(*ptr && IsNum(*ptr)))
		{
			ptr--;
			return res;
		}

		while (*ptr && IsNum(*ptr))
		{
			res = 10 * res + (*ptr - '0');
			ptr++;
		}

		return -res;
	}
	else
	{
		while (*ptr && IsNum(*ptr))
		{
			res = 10 * res + (*ptr - '0');
			ptr++;
		}

		return res;
	}
}

string Scenario::ReadSymbol(void)
{
	string res("");
	if (!(*ptr && IsAlpha(*ptr))) return res;

	do {
		res += *(ptr++);
	} while ((*ptr) && IsAlphaNum(*ptr));

	return res;
}

string Scenario::ReadString(void)
{
	string res("");
	char* bptr = ptr;

	if (*ptr != '"') return res;
	ptr++;

	while (*ptr)
	{
		if (*ptr == '"')
		{
			ptr++;

			if (*ptr == '"')
			{
				ptr++;
				res += '"';
				continue;
			}

			return res;
		}

		if (*ptr=='-' && ptr[1]=='-' && ptr[2]=='-')
		{
			ptr += 3;
			res += '\n';
			continue;
		}

		res += *ptr;
		ptr++;
	}

	ptr = bptr;
	return res;
}

Scenario::Expression Scenario::StrValue(Expression expr)
{
	if (expr.type == EXPR_STR_VALUE) return expr;

	Expression res;
	res.type = EXPR_FUNC;
	res.strValue = string("str");
	res.params.push_back(expr);

	return res;
}

Scenario::Expression Scenario::GenerateAddExpr(vector<Expression>& lst, int cnt)
{
	Expression res;
	res.type = EXPR_OP;
	res.op = '+';

	res.params.push_back(StrValue(lst[cnt]));

	if (cnt+2 < lst.size()) res.params.push_back(GenerateAddExpr(lst, cnt+1));
	else res.params.push_back(StrValue(lst[cnt+1]));

	return res;
}

Scenario::Parameter Scenario::ReadParamString(const char* path, int lineNum)
{
	string buf("");
	bool err = true;
	vector<Expression> lst;

	while (*ptr)
	{
		if (*ptr == '"')
		{
			ptr++;

			if (*ptr == '"')
			{
				ptr++;
				buf += '"';
				continue;
			}

			if (!buf.empty())
			{
				Expression expr;
				expr.type = EXPR_STR_VALUE;
				expr.strValue = buf;
				lst.push_back(expr);
			}

			err = false;
			break;
		}

		if (*ptr=='-' && ptr[1]=='-' && ptr[2]=='-')
		{
			ptr += 3;
			buf += '\n';
			continue;
		}

		if (*ptr == '#')
		{
			ptr++;

			if (*ptr == '#')
			{
				ptr++;
				buf += '#';
				continue;
			}

			if (!buf.empty())
			{
				Expression expr;
				expr.type = EXPR_STR_VALUE;
				expr.strValue = buf;
				lst.push_back(expr);
				buf = string("");
			}

			if (*ptr == '{')
			{
				ptr++;
				SkipSpaces();

				lst.push_back(ReadExpression(path, lineNum));
				SkipSpaces();

				if (*ptr != '}') throw Exception("Invalid eval expression in %s at line %d", path, lineNum);
				ptr++;
			}
			else lst.push_back(ReadExpression(path, lineNum));

			continue;
		}

		buf += *ptr;
		ptr++;
	}

	if (err) throw Exception("Invalid parameter string in %s at line %d", path, lineNum);

	Parameter par;

	if (lst.size() == 0)
	{
		par.type = PARAM_STRING;
		par.strValue = string("");
	}
	else if (lst.size()==1 && lst[0].type==EXPR_STR_VALUE)
	{
		par.type = PARAM_STRING;
		par.strValue = lst[0].strValue;
	}
	else if (lst.size() == 1)
	{
		par.type = PARAM_EXPRESSION;
		par.expr = lst[0];
	}
	else
	{
		par.type = PARAM_EXPRESSION;
		par.expr = GenerateAddExpr(lst, 0);
	}

	return par;
}

void Scenario::ReadParams(list<Parameter>& params, const char* path, int lineNum)
{
	string name("");

	while (*ptr)
	{
		if (*ptr == '#')
		{
			ptr++;
			if (!(*ptr)) throw Exception("Missing eval expression in %s at line %d", path, lineNum);

			Parameter par;
			par.type = PARAM_EXPRESSION;
			par.name = name;

			if (*ptr == '{')
			{
				ptr++;
				SkipSpaces();

				par.expr = ReadExpression(path, lineNum);
				SkipSpaces();

				if (*ptr != '}') throw Exception("Invalid eval expression in %s at line %d", path, lineNum);
				ptr++;
			}
			else par.expr = ReadExpression(path, lineNum);

			SkipSpaces();
			params.push_back(par);
			name = "";
		}
		else
		if (*ptr == '"')
		{
			ptr++;
			Parameter par = ReadParamString(path, lineNum);
			par.name = name;
			params.push_back(par);
			name = "";
		}
		else
		if (IsNum(*ptr) || *ptr=='-')
		{
			char *bptr = ptr;
			int num = ReadNumber();
			if (*ptr || ptr==bptr) throw Exception("Invalid parameter number in %s at line %d", path, lineNum);

			Parameter par;
			par.type = PARAM_NUMBER;
			par.name = name;
			par.numValue = num;
			params.push_back(par);

			name = "";
		}
		else
		{
			string buf = ReadSymbol();
			if (buf.empty()) throw Exception("Invalid parameter symbol in %s at line %d", path, lineNum);

			SkipSpaces();

			if (*ptr == '=')
			{
				if (!name.empty()) throw Exception("Double named parameter in %s at line %d", path, lineNum);
				name = buf;
				ptr++;
				SkipSpaces();
				continue;
			}

			Parameter par;
			par.type = PARAM_SYMBOL;
			par.name = name;
			par.strValue = buf;
			params.push_back(par);

			name = "";
		}

		if (*ptr)
		{
			SkipSpaces();
			if (!(*ptr)) throw Exception("Extra spaces in %s at line %d", path, lineNum);
			if (*ptr != ',') throw Exception("Invalid parameters syntax in %s at line %d", path, lineNum);

			ptr++;
			SkipSpaces();

			if (!(*ptr)) throw Exception("Extra spaces in %s at line %d", path, lineNum);
		}
	}
}

Scenario::Parameter Scenario::ReadOneParam(const char* path, int lineNum)
{
	Parameter par;
	par.name = string("");

	if (*ptr == '#')
	{
		ptr++;
		if (!(*ptr)) throw Exception("Missing eval expression in %s at line %d", path, lineNum);

		par.type = PARAM_EXPRESSION;

		if (*ptr == '{')
		{
			ptr++;
			SkipSpaces();

			par.expr = ReadExpression(path, lineNum);
			SkipSpaces();

			if (*ptr != '}') throw Exception("Invalid eval expression in %s at line %d", path, lineNum);
			ptr++;
		}
		else par.expr = ReadExpression(path, lineNum);
	}
	else
	if (*ptr == '"')
	{
		ptr++;
		par = ReadParamString(path, lineNum);
		par.name = string("");
	}
	else
	if (IsNum(*ptr) || *ptr=='-')
	{
		char *bptr = ptr;
		int num = ReadNumber();
		if (*ptr || ptr==bptr) throw Exception("Invalid parameter number in %s at line %d", path, lineNum);

		par.type = PARAM_NUMBER;
		par.numValue = num;
	}
	else
	{
		string buf = ReadSymbol();
		if (buf.empty()) throw Exception("Invalid parameter symbol in %s at line %d", path, lineNum);

		par.type = PARAM_SYMBOL;
		par.strValue = buf;
	}

	SkipSpaces();
	return par;
}

// Precedence climbing
// http://www.engr.mun.ca/~theo/Misc/exp_parsing.htm

int Scenario::ReadExpressionPrec(int op)
{
	switch (op)
	{
		case '|':
			return 0;

		case '&':
			return 1;

		case '=':
		case '>':
		case '<':
		case OP_NEQ:
		case OP_LTEQ:
		case OP_GTEQ:
			return 2;

		case '+':
		case '-':
			return 3;

		case '*':
		case '/':
		case '%':
			return 5;
	}
}

int Scenario::ReadExpressionPrecUn(int op)
{
	// op can be '-' or '!'
	return 4;
}

Scenario::Expression Scenario::ReadExpressionNext(bool first)
{
	Expression res;

	if (!(*ptr))
	{
		res.type = EXPR_END;
	}
	if (first && (*ptr)=='-')
	{
		res.type = EXPR_OP;
		res.op = OP_NEG;
	}
	else if (first && (*ptr)=='!')
	{
		res.type = EXPR_OP;
		res.op = OP_NOT;
	}
	else if ((*ptr)=='+' || (*ptr)=='-' || (*ptr)=='*' || (*ptr)=='/' || (*ptr)=='%' || (*ptr)=='|' || (*ptr)=='&' || (*ptr)=='=')
	{
		res.type = EXPR_OP;
		res.op = *ptr;
	}
	else if (*ptr == '>')
	{
		res.type = EXPR_OP;

		if (ptr[1] == '=')
		{
			res.op = OP_GTEQ;
			ptr++;
		}
		else res.op = '>';
	}
	else if (*ptr == '<')
	{
		res.type = EXPR_OP;

		if (ptr[1] == '=')
		{
			res.op = OP_LTEQ;
			ptr++;
		}
		else res.op = '<';
	}
	else if (*ptr == '!')
	{
		if (ptr[1] != '=')
		{
			res.type = EXPR_END;
			return res;
		}

		res.type = EXPR_OP;
		res.op = OP_NEQ;
	}
	else if (*ptr == '(')
	{
		res.type = EXPR_BRAC;
	}
	else if (*ptr == '"')
	{
		char* bptr = ptr;
		res.strValue = ReadString();

		if (ptr == bptr)
		{
			res.type = EXPR_END;
			return res;
		}

		res.type = EXPR_STR_VALUE;
		SkipSpaces();
		return res;
	}
	else if (IsNum(*ptr))
	{
		char *bptr = ptr;
		res.numValue = ReadNumber();

		if (ptr == bptr)
		{
			res.type = EXPR_END;
			return res;
		}

		res.type = EXPR_NUM_VALUE;
		SkipSpaces();
		return res;
	}
	else
	{
		res.strValue = ReadSymbol();

		if (res.strValue.empty())
		{
			res.type = EXPR_END;
			return res;
		}

		res.type = EXPR_SYMBOL;
		SkipSpaces();
		return res;
	}

	ptr++;
	SkipSpaces();
	return res;
}

Scenario::Expression Scenario::ReadExpressionP(bool first, const char* path, int lineNum)
{
	Expression res;
	Expression next = ReadExpressionNext(first);

	if (next.type==EXPR_OP && (next.op==OP_NOT || next.op==OP_NEG))
	{
		res.type = EXPR_OP;
		res.op = next.op;
		res.params.push_back(ReadExpressionExp(ReadExpressionPrecUn(next.op), false, path, lineNum));
	}
	else if (next.type == EXPR_BRAC)
	{
		res = ReadExpressionExp(0, true, path, lineNum);
		if (*ptr != ')') throw Exception("Invalid expression (1) in %s at line %d", path, lineNum);
		ptr++;
		SkipSpaces();
	}
	else if (next.type==EXPR_NUM_VALUE || next.type==EXPR_STR_VALUE)
	{
		return next;
	}
	else if (next.type == EXPR_SYMBOL)
	{
		res.type = EXPR_FUNC;
		res.strValue = next.strValue;

		if (*ptr == '(')
		{
			if (compileTimeWarn) {
				if (mathFunctions.find(res.strValue) == mathFunctions.end()) {
					throw Exception("Undefined math function \"%s\" in %s at line %d", res.strValue.c_str(), path, lineNum);
				}
			}

			ptr++;
			SkipSpaces();

			if (*ptr != ')')
			{
				while (*ptr)
				{
					res.params.push_back(ReadExpressionExp(0, true, path, lineNum));

					if (*ptr == ')') break;
					if (*ptr != ',') throw Exception("Invalid expression (2) in %s at line %d", path, lineNum);

					ptr++;
					SkipSpaces();
				}
	
				if (!(*ptr)) throw Exception("Invalid expression (3) in %s at line %d", path, lineNum);
			}

			ptr++;
			SkipSpaces();
		}
		else
		{
			res.type = EXPR_VAR;
			res.strValue = next.strValue;
		}
	}
	else throw Exception("Invalid expression (4) in %s at line %d", path, lineNum);

	return res;
}

Scenario::Expression Scenario::ReadExpressionExp(int p, bool first, const char* path, int lineNum)
{
	Expression res = ReadExpressionP(first, path, lineNum);

	char* bptr = ptr;
	Expression next = ReadExpressionNext(false);

	while (next.type==EXPR_OP && ReadExpressionPrec(next.op)>=p)
	{
		Expression nres;
		nres.type = EXPR_OP;
		nres.op = next.op;
		nres.params.push_back(res);
		nres.params.push_back(ReadExpressionExp(ReadExpressionPrec(next.op) + 1, false, path, lineNum));

		res = nres;
		bptr = ptr;
		next = ReadExpressionNext(false);
	}

	ptr = bptr;
	return res;
}

Scenario::Expression Scenario::ReadExpression(const char* path, int lineNum)
{
	if (!(*ptr)) throw Exception("No expression found in %s at line %d", path, lineNum);
	Expression res = ReadExpressionExp(0, true, path, lineNum);
	return res;
}

void Scenario::Load(const char* path)
{
	File fl;

	FreeScript();
	fl.Read(path);

	int lineNum = 0;
	scriptPath = path;

	while (!fl.Eof())
	{
		lineNum++;
		fl.GetS(buffer, 0x1000);

		ptr = buffer;
		SkipSpaces();

		if (!(*ptr) || *ptr=='\'') continue;

		if (*ptr == ':')
		{
			ptr++;
			string label = ReadSymbol();

			if (*ptr || label=="") throw Exception("Invalid label in %s at line %d", path, lineNum);
			if (labels.find(label) != labels.end()) throw Exception("Label %s already defined", label.c_str());

			labels[label] = script.size();
			continue;
		}

		string buf = ReadSymbol();
		if (buf.empty()) throw Exception("Syntax error in %s at line %d", path, lineNum);
		SkipSpaces();

		if (*ptr == '=')
		{
			ptr++;
			SkipSpaces();

			Line ln;
			ln.type = OPERATOR_SET;
			ln.lineNum = lineNum;
			ln.action = new OperatorSet;
			((OperatorSet*)ln.action)->variable = buf;
			((OperatorSet*)ln.action)->expr = ReadExpression(path, lineNum);

			script.push_back(ln);
			continue;
		}

		if (buf == string("if"))
		{
			Expression expr = ReadExpression(path, lineNum);
			SkipSpaces();

			Line ln;
			ln.type = OPERATOR_IF;
			ln.lineNum = lineNum;
			ln.action = new OperatorIf;
			((OperatorIf*)ln.action)->cond = expr;
			((OperatorIf*)ln.action)->labelParam = ReadOneParam(path, lineNum);

			script.push_back(ln);
			continue;
		}

		if (compileTimeWarn) {
			if (functions.find(buf) == functions.end()) {
				throw Exception("Engine function \"%s\" not found in %s at line %d", buf.c_str(), path, lineNum);
			}
		}

		Line lne;
		lne.type = OPERATOR_EXEC;
		lne.lineNum = lineNum;
		lne.action = new OperatorExec;
		((OperatorExec*)lne.action)->action = buf;
		ReadParams(((OperatorExec*)lne.action)->params, path, lineNum);
		script.push_back(lne);
	}

	status = SCRIPT_LOADED;
	fl.Close();
}

void Scenario::DefineFunc(const char* name, EngineFunc func)
{
	if (functions.find(string(name)) != functions.end()) throw Exception("Engine function \"%s\" already defined", name);
	functions[string(name)] = func;
}

void Scenario::DefineMathFunc(const char* name, MathFunc func)
{
	if (mathFunctions.find(string(name)) != mathFunctions.end()) throw Exception("Math function \"%s\" already defined", name);
	mathFunctions[string(name)] = func;
}

void Scenario::Init(void)
{
	if (status < SCRIPT_LOADED) throw Exception("Load script before calling Init");

	currentLine = 0;
	variables.clear();
	status = SCRIPT_RUN;
}

Scenario::Variable Scenario::CalcExpression(Expression& expr, int lineNum)
{
	Variable res;

	if (expr.type == EXPR_NUM_VALUE)
	{
		res.type = VAR_NUMBER;
		res.numValue = expr.numValue;
		return res;
	}

	if (expr.type == EXPR_STR_VALUE)
	{
		res.type = VAR_STRING;
		res.strValue = expr.strValue;
		return res;
	}

	if (expr.type==EXPR_OP && expr.op==OP_NOT)
	{
		Variable va = CalcExpression(expr.params[0], lineNum);
		res.type = VAR_NUMBER;

		if (va.type == VAR_NUMBER) res.numValue = (va.numValue == 0 ? 1 : 0);
		else res.numValue = !res.strValue.empty();

		return res;
	}

	if (expr.type==EXPR_OP && expr.op==OP_NEG)
	{
		Variable va = CalcExpression(expr.params[0], lineNum);
		if (va.type != VAR_NUMBER) throw Exception("Invalid operand type in %s at line %d", scriptPath.c_str(), lineNum);

		res.type = VAR_NUMBER;
		res.numValue = -va.numValue;
		return res;
	}

	if (expr.type == EXPR_OP)
	{
		Variable va = CalcExpression(expr.params[0], lineNum);
		Variable vb = CalcExpression(expr.params[1], lineNum);

		if (va.type != vb.type) throw Exception("Defferent operand types in %s at line %d", scriptPath.c_str(), lineNum);

		switch (expr.op)
		{
			case '+':
				if (va.type == VAR_NUMBER) {
					res.type = VAR_NUMBER;
					res.numValue = va.numValue + vb.numValue;
				} else {
					res.type = VAR_STRING;
					res.strValue = va.strValue + vb.strValue;
				}
				break;

			case '-':
				if (va.type != VAR_NUMBER) throw Exception("Invalid operands types in %s at line %d", scriptPath.c_str(), lineNum);
				res.type = VAR_NUMBER;
				res.numValue = va.numValue - vb.numValue;
				break;

			case '*':
				if (va.type != VAR_NUMBER) throw Exception("Invalid operands types in %s at line %d", scriptPath.c_str(), lineNum);
				res.type = VAR_NUMBER;
				res.numValue = va.numValue * vb.numValue;
				break;

			case '/':
				if (va.type != VAR_NUMBER) throw Exception("Invalid operands types in %s at line %d", scriptPath.c_str(), lineNum);
				if (vb.numValue == 0) throw Exception("Divizion by zero in %s at line %d", scriptPath.c_str(), lineNum);
				res.type = VAR_NUMBER;
				res.numValue = va.numValue / vb.numValue;
				break;

			case '%':
				if (va.type != VAR_NUMBER) throw Exception("Invalid operands types in %s at line %d", scriptPath.c_str(), lineNum);
				if (vb.numValue == 0) throw Exception("Divizion by zero in %s at line %d", scriptPath.c_str(), lineNum);
				res.type = VAR_NUMBER;
				res.numValue = va.numValue % vb.numValue;
				break;

			case '>':
				res.type = VAR_NUMBER;
				res.numValue = (va.type==VAR_NUMBER ? (va.numValue > vb.numValue) : (va.strValue > vb.strValue));
				break;

			case '<':
				res.type = VAR_NUMBER;
				res.numValue = (va.type==VAR_NUMBER ? (va.numValue < vb.numValue) : (va.strValue < vb.strValue));
				break;

			case '=':
				res.type = VAR_NUMBER;
				res.numValue = (va.type==VAR_NUMBER ? (va.numValue == vb.numValue) : (va.strValue == vb.strValue));
				break;

			case OP_LTEQ:
				res.type = VAR_NUMBER;
				res.numValue = (va.type==VAR_NUMBER ? (va.numValue <= vb.numValue) : (va.strValue <= vb.strValue));
				break;

			case OP_GTEQ:
				res.type = VAR_NUMBER;
				res.numValue = (va.type==VAR_NUMBER ? (va.numValue >= vb.numValue) : (va.strValue >= vb.strValue));
				break;

			case OP_NEQ:
				res.type = VAR_NUMBER;
				res.numValue = (va.type==VAR_NUMBER ? (va.numValue != vb.numValue) : (va.strValue != vb.strValue));
				break;

			case '&':
				res.type = VAR_NUMBER;
				res.numValue = (va.type==VAR_NUMBER ? (va.numValue && vb.numValue) : (!va.strValue.empty() && !vb.strValue.empty()));
				break;

			case '|':
				res.type = VAR_NUMBER;
				res.numValue = (va.type==VAR_NUMBER ? (va.numValue && vb.numValue) : (!va.strValue.empty() || !vb.strValue.empty()));
				break;
		}

		return res;
	}

	if (expr.type == EXPR_VAR)
	{
		if (variables.find(expr.strValue) == variables.end())
		{
			if (undefinedVariableException) throw Exception("Undefined variable %s in %s at line %d", expr.strValue.c_str(), scriptPath.c_str(), lineNum);

			Variable vr;
			vr.type = VAR_NUMBER;
			vr.numValue = 0;
 			return vr;
		}

		return variables[expr.strValue];
	}

	if (mathFunctions.find(expr.strValue) == mathFunctions.end()) throw Exception("Undefined math function \"%s\" in %s at line %d", expr.strValue.c_str(), scriptPath.c_str(), lineNum);

	vector<Variable> params;
	for (vector<Expression>::iterator it = expr.params.begin(); it != expr.params.end(); it++) params.push_back(CalcExpression(*it, lineNum));

	MathFunc* func = &mathFunctions[expr.strValue];
	return func->func(params, this, lineNum);
}

bool Scenario::CalcCondition(Expression& expr, int lineNum)
{
	Variable var = CalcExpression(expr, lineNum);
	return ((var.type == VAR_NUMBER) ? (var.numValue != 0) : (!var.strValue.empty()));
}

void Scenario::GotoLabel(string label, int lineNum)
{
	if (labels.find(label) == labels.end()) throw Exception("Undefined label %s in %s at line %d", label.c_str(), scriptPath.c_str(), lineNum);
	if (infLoop.find(label)!=infLoop.end() && !varsChanged) throw Exception("Found infinitive loop in %s at line %d", scriptPath.c_str(), lineNum);

	varsChanged = false;
	infLoop[label] = true;
	currentLine = labels[label];
}

void Scenario::ProcessIf(Line* ln)
{
	if (!CalcCondition(((OperatorIf*)ln->action)->cond, ln->lineNum)) return;

	Parameter par = EvalParam(((OperatorIf*)ln->action)->labelParam, ln->lineNum);
	if (par.type!=PARAM_SYMBOL && par.type!=PARAM_STRING) throw Exception("Invalid label in %s at line %d", scriptPath.c_str(), ln->lineNum);

	GotoLabel(par.strValue, ln->lineNum);
}

Scenario::Parameter Scenario::FindParamByName(list<Parameter>* params, string name)
{
	list<Parameter>::iterator it;

	for (it = params->begin(); it != params->end(); it++) {
		if (it->name == name) return *it;
	}

	Parameter none;
	none.type = PARAM_NONE;
	return none;
}

Scenario::Parameter Scenario::EvalParam(Parameter param, int lineNum)
{
	if (param.type != PARAM_EXPRESSION) return param;

	Parameter res;
	res.name = param.name;

	Variable var = CalcExpression(param.expr, lineNum);

	if (var.type == VAR_NUMBER)
	{
		res.type = PARAM_NUMBER;
		res.numValue = var.numValue;
	}
	else
	{
		res.type = PARAM_STRING;
		res.strValue = var.strValue;
	}

	return res;
}

void Scenario::ProcessFunc(Line* ln)
{
	string* funcName = &(((OperatorExec*)ln->action)->action);

	if (functions.find(*funcName) == functions.end()) throw Exception("Engine function \"%s\" not found in %s at line %d", funcName->c_str(), scriptPath.c_str(), ln->lineNum);

	EngineFunc* func = &functions[*funcName];
	list<Parameter>* fparams = &(((OperatorExec*)ln->action)->params);
	hash_map<string, Parameter> params;

	list<string>::iterator itp;
	list<Parameter>::iterator itf;

	for (itp = func->paramNames.begin(), itf = fparams->begin(); itp != func->paramNames.end(); itp++)
	{
		if (itf != fparams->end())
		{
			if (itf->name.empty())
			{
				params[itp->c_str()] = EvalParam(*itf, ln->lineNum);
				itf++;
			}
			else
			{
				itf = fparams->end();
				params[itp->c_str()] = EvalParam(FindParamByName(fparams, *itp), ln->lineNum);
			}
		}
		else params[itp->c_str()] = EvalParam(FindParamByName(fparams, *itp), ln->lineNum);
	}

	func->func(params, this, ln->lineNum);
}

void Scenario::NextLine(void)
{
	if (status == SCRIPT_END) return;

	if (currentLine >= script.size())
	{
		status = SCRIPT_END;
		return;
	}

	Line* ln = &script[currentLine];
	currentLine++;

	switch (ln->type)
	{
		case OPERATOR_SET:
			variables[((OperatorSet*)ln->action)->variable] = CalcExpression(((OperatorSet*)ln->action)->expr, ln->lineNum);
			varsChanged = true;
			break;

		case OPERATOR_IF:
			ProcessIf(ln);
			break;

		case OPERATOR_EXEC:
			ProcessFunc(ln);
			break;
	}
}

void Scenario::Step(void)
{
	if (status < SCRIPT_RUN) throw Exception("Call Init before calling Step");
	if (status == SCRIPT_END) return;

	varsChanged = false;
	infLoop.clear();
	NextLine();
}

int Scenario::GetStatus(void)
{
	return status;
}

void Scenario::SetStatusEnd(void)
{
	if (status < SCRIPT_RUN) throw Exception("Call Init before calling SetStatusEnd");
	status = SCRIPT_END;
}

void Scenario::FunctionGoto(hash_map<string, Parameter> params, Scenario* scen, int lineNum)
{
	if (params[string("label")].type == PARAM_NONE) throw Exception("Undefined label in %s at line %d", scen->scriptPath.c_str(), lineNum);
	if (params[string("label")].type!=PARAM_SYMBOL && params[string("label")].type!=PARAM_STRING) throw Exception("Invalid label in %s at line %d", scen->scriptPath.c_str(), lineNum);

	scen->GotoLabel(params[string("label")].strValue, lineNum);
}

Scenario::Variable Scenario::MathStr(vector<Variable> params, Scenario* scen, int lineNum)
{
	if (params.size() != 1) throw Exception("Invalid params count in \"str\" math function in %s at line %d", scen->scriptPath.c_str(), lineNum);

	Variable var = params[0];

	if (var.type != VAR_STRING)
	{
		char buf[0x100];
		sprintf(buf, "%d", var.numValue);

		var.type = VAR_STRING;
		var.strValue = string(buf);
	}

	return var;
}

Scenario::Variable Scenario::MathRand(vector<Variable> params, Scenario* scen, int lineNum)
{
	if (params.size() != 1) throw Exception("Invalid params count in \"rand\" math function in %s at line %d", scen->scriptPath.c_str(), lineNum);

	Variable var = params[0];
	if (var.type != VAR_NUMBER) throw Exception("Invalid parameter type in \"rand\" math function in %s at line %d", scen->scriptPath.c_str(), lineNum);

	var.numValue = (rand() % var.numValue);
	return var;
}
