#ifndef __SCENARIO_H__
#define __SCENARIO_H__

#include <vector>
#include <ext/hash_map>
#include <string>
#include <list>

#include "file.h"
#include "dict.h"

using namespace std;
using namespace ext;

#define VAR_NUMBER 1
#define VAR_STRING 2

#define OP_NOT 0x100
#define OP_NEG 0x101
#define OP_LTEQ 0x102
#define OP_GTEQ 0x103
#define OP_NEQ 0x104

#define EXPR_NUM_VALUE 1
#define EXPR_STR_VALUE 2
#define EXPR_OP 3
#define EXPR_VAR 4
#define EXPR_FUNC 5
#define EXPR_BRAC 10
#define EXPR_SYMBOL 11
#define EXPR_END 12

#define OPERATOR_SET 1
#define OPERATOR_IF 2
#define OPERATOR_EXEC 3

#define PARAM_NONE 0
#define PARAM_STRING 1
#define PARAM_SYMBOL 2
#define PARAM_NUMBER 3
#define PARAM_EXPRESSION 4

#define SCRIPT_NONE 0
#define SCRIPT_LOADED 1
#define SCRIPT_RUN 2
#define SCRIPT_END 3

class Scenario
{
	public:

	struct Variable
	{
		int type;
		int numValue;
		string strValue;
	};

	struct Expression
	{
		int type;

		int op;
		int numValue;
		string strValue;

		vector<Expression> params;
	};

	struct Parameter
	{
		int type;
		string name;

		int numValue;
		string strValue;
		Expression expr;		// don't use it in EngineFunc, because expression wil be calculated by script engine
	};

	struct OperatorSet
	{
		string variable;
		Expression expr;
	};

	struct OperatorIf
	{
		Expression cond;
		Parameter labelParam;
	};

	struct OperatorExec
	{
		string action;
		list<Parameter> params;
	};

	struct Line
	{
		int type;
		int lineNum;
		void* action;
	};

	struct EngineFunc
	{
		void (* func)(hash_map<string, Parameter>, Scenario*, int);
		list<string> paramNames;
	};

	struct MathFunc
	{
		Variable (* func)(vector<Variable> params, Scenario*, int);
	};

	string scriptPath;
	vector<Line> script;
	hash_map<string, int> labels;
	hash_map<string, Variable> variables;
	hash_map<string, EngineFunc> functions;
	hash_map<string, MathFunc> mathFunctions;

	Scenario();
	Scenario(bool compileTimeWarn, bool undefinedVariableException);
	~Scenario();

	Scenario(const Scenario& s);
	Scenario& operator=(const Scenario& s);

	void Load(const char* path);
	void DefineFunc(const char* name, EngineFunc func);
	void DefineMathFunc(const char* name, MathFunc func);

	void Init(void);
	void Step(void);

	void GotoLabel(string label, int lineNum);

	int GetStatus(void);
	void SetStatusEnd(void);

	protected:

	bool compileTimeWarn;
	bool undefinedVariableException;

	int status;
	int currentLine;
	bool varsChanged;
	hash_map<string, bool> infLoop;

	void Construct(void);

	Variable CalcExpression(Expression& expr, int lineNum);
	bool CalcCondition(Expression& expr, int lineNum);

	void NextLine(void);
	void ProcessIf(Line* ln);
	void ProcessFunc(Line* ln);
	Parameter EvalParam(Parameter param, int lineNum);
	Parameter FindParamByName(list<Parameter>* params, string name);

	static void FunctionGoto(hash_map<string, Parameter> params, Scenario* scen, int lineNum);
	static Variable MathStr(vector<Variable> params, Scenario* scen, int lineNum);
	static Variable MathRand(vector<Variable> params, Scenario* scen, int lineNum);

	char *ptr;
	char buffer[0x1000];

	bool IsNum(char c);
	bool IsAlpha(char c);
	bool IsAlphaNum(char c);

	void SkipSpaces(void);

	int ReadNumber(void);
	string ReadSymbol(void);
	string ReadString(void);

	Expression StrValue(Expression expr);
	Expression GenerateAddExpr(vector<Expression>& lst, int cnt);
	Parameter ReadParamString(const char* path, int lineNum);
	void ReadParams(list<Parameter>&, const char* path, int lineNum);
	Parameter ReadOneParam(const char* path, int lineNum);

	int ReadExpressionPrec(int op);
	int ReadExpressionPrecUn(int op);
	Expression ReadExpressionNext(bool first);
	Expression ReadExpressionP(bool first, const char* path, int lineNum);
	Expression ReadExpressionExp(int p, bool first, const char* path, int lineNum);
	Expression ReadExpression(const char* path, int lineNum);

	void FreeScript(void);
};

#endif /* __SCENARIO_H__ */
